"""High-level API for files from fileman service."""
import math
from array import array
from datetime import datetime
from io import SEEK_END, SEEK_SET, BufferedIOBase, BytesIO, UnsupportedOperation
from itertools import repeat, starmap
from mmap import mmap
from typing import Optional, Union, cast
from uuid import UUID

from .client import FileClient

# Set the defalt buffer to 3 MB.
# The default limit for gRPC message is 4 MB. It needs to contain the file data itself and the rest of message.
DEFAULT_BUFFER_SIZE = 3 * 1024 ** 2


class File(BufferedIOBase):
    """Provide complete Binary I/O for a file in a fileman service.

    Raises:
        UnsupportedOperation: If the required operation is not supported in current mode.
    """

    def __init__(self, client: FileClient, *, uuid: UUID = None, name: str = None, mimetype: str = None):
        self.client = client
        self._uuid = uuid
        if self._write_mode:
            if name is None:
                raise TypeError("One of uuid and name must be provided.")
            # Leave the metadata empty, but initialize the internal buffer.
            # TODO: Use TypedDict for annotations.
            self._stat = {'name': name, 'mimetype': mimetype}
            self._buff: Optional[BytesIO] = None
        else:
            if name is not None:
                raise TypeError("uuid and name can't be both provided.")
            if mimetype is not None:
                raise TypeError("uuid and mimetype can't be both provided.")
            # Fetch metadata on read mode and leave the internal buffer empty.
            self._stat = self.client.stat(cast(UUID, self._uuid))
            self._buff = None

    @property
    def _write_mode(self) -> bool:
        """Return whether the file is in a write mode."""
        return self._uuid is None

    @property
    def uuid(self) -> Optional[UUID]:
        """Return file's UUID."""
        return self._uuid

    @property
    def create_datetime(self) -> datetime:
        """Return file's create datetime."""
        try:
            return cast(datetime, self._stat['create_datetime'])
        except KeyError:
            raise UnsupportedOperation("create_datetime is not supported in write mode.")

    @property
    def size(self) -> int:
        """Return file's size."""
        try:
            return cast(int, self._stat['size'])
        except KeyError:
            raise UnsupportedOperation("size is not supported in write mode.")

    @property
    def mimetype(self) -> Optional[str]:
        """Return file's MIME type."""
        return cast(str, self._stat['mimetype'])

    @property
    def name(self) -> str:
        """Return file's name."""
        return cast(str, self._stat['name'])

    @property
    def _raw(self) -> BytesIO:
        """Return the internal buffer. Create it if needed."""
        if self._buff is None:
            if self._write_mode:
                # Return a one time buffer to mock the API.
                return BytesIO()
            else:
                # Fetch the file.
                self._buff = BytesIO(b''.join(self.client.read(cast(UUID, self.uuid), DEFAULT_BUFFER_SIZE)))
        return self._buff

    # Binary I/O API - IOBase https://docs.python.org/3/library/io.html#io.IOBase
    @property
    def closed(self) -> bool:
        """Return `False`, this file is never closed."""
        return False

    def flush(self) -> None:
        """Flush write buffer, if applicable.

        This actually creates the file in fileman.
        This is a noop if in read mode or if there is nothing to be written.
        Also ends the write mode on the file.
        """
        if self._write_mode and self._buff:
            # Prepare chunks.
            # Duplicate buffer to prevent unwated modification of self._buff.
            buff = BytesIO(self._buff.getvalue())
            buff_len = buff.seek(0, SEEK_END)
            buff.seek(0)
            num_chunks = max(math.ceil(buff_len / DEFAULT_BUFFER_SIZE), 1)
            # Based on 'repeatfunc' recipe from itertools.
            chunks = starmap(buff.read, repeat((DEFAULT_BUFFER_SIZE, ), num_chunks))

            self._uuid = self.client.create(self.name, chunks, mimetype=self.mimetype)
            # Fetch complete metadata
            self._stat = self.client.stat(self._uuid)

    def readable(self) -> bool:
        """Return `True`, this stream can be read from."""
        return True

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        """Change the stream position to the given byte offset."""
        return self._raw.seek(offset, whence)

    def seekable(self) -> bool:
        """Return `True`, this stream support random access."""
        return True

    def writable(self) -> bool:
        """Return `True` if this file is in a write mode."""
        return self._write_mode

    # Binary I/O API - BufferedIOBase https://docs.python.org/3/library/io.html#io.BufferedIOBase

    def read(self, size: Optional[int] = -1) -> bytes:
        """Read and return up to size bytes.

        If the argument is omitted, None, or negative, data is read and returned until EOF is reached.
        """
        return self._raw.read(size)

    def write(self, b: Union[bytes, bytearray, memoryview, array, mmap]) -> int:  # type: ignore[override] # _CData
        """Write data into an internal buffer."""
        if not self._write_mode:
            raise UnsupportedOperation("File is not in a write mode.")
        if not self._buff:
            self._buff = BytesIO()
        return self._buff.write(b)


class Storage:
    """Utility storage to retrieve files from single service."""

    def __init__(self, client: FileClient):
        self.client = client

    def open(self, uuid: UUID) -> File:
        """Return a file with corresponding UUID."""
        return File(self.client, uuid=uuid)

    def create(self, name: str, *, mimetype: str = None) -> File:
        """Create a new file."""
        return File(self.client, name=name, mimetype=mimetype)
